/*
 *	wave.c
 *
 *  Copyright (C) 1998 Hermann Samso
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ncurses.h>
#include <string.h>
#include "delfin.h"
#include "wave.h"
#include "display.h"
#include "file.h"
#include "audiodrv.h"

INT WaveIdxByName(CHAR* name);
void WaveSwap(INT sampleidx);

WAVE* waves[MAXWAVES];			// wave pool
INT nrwaves = 0;				// # of active waves < MAXWAVES
INT idx = 0;					// index of selected wave
// samples out of range should be cutted?
BOOL left_up_cut = TRUE;
BOOL left_down_cut = TRUE;
BOOL right_up_cut = TRUE;
BOOL right_down_cut = TRUE;

/* 
 * Creates a WAVE from a .raw 16bit, 44100Hz file
 * 'channels' == 1 mono, == 2 stereo
 */
WAVE* WaveLoadRaw(CHAR* fname, int channels) {
	ULONG raw_bytes;
	ULONG raw_words;
	SWORD* raw;
	WAVE* wave = NULL;
	ULONG i;
	
		
	if (trace_on)
		fprintf(stderr, "WaveLoadRaw(\"%s\")\n", fname);


	raw_bytes = FileSize(fname);
	raw_words = raw_bytes/2;
	raw = (SWORD*)FileReadAll(fname);
	if (raw == NULL) {
		fprintf(stderr, "WaveLoadRaw: couldn't read file!\n");
		return NULL;
	}

	if (channels == 1) {
	// it's a mono wave
		wave = WaveNew(fname, raw_words);
		for (i = 0; i < wave->size; i=i+1) {
			wave->samples[i].left = (FLOAT)raw[i]/32767;
			wave->samples[i].right = wave->samples[i].left;
			/* calc peak */
			if (wave->samples[i].left > wave->peak)
				wave->peak = wave->samples[i].left;
		}
	}
	else {
	// it's a stereo wave
		wave = WaveNew(fname, raw_words/2);
		for (i = 0; i < wave->size; i=i+1) {
			wave->samples[i].left = (FLOAT)raw[i*2]/32767;
			wave->samples[i].right = (FLOAT)raw[i*2+1]/32767;
			/* calc peak */
			if (wave->samples[i].left > wave->peak)
				wave->peak = wave->samples[i].left;
			if (wave->samples[i].right > wave->peak)
				wave->peak = wave->samples[i].right;
		}
	}

	WaveRegister(wave);
	return(wave); 
}


void WaveSaveRaw(CHAR* fname, WAVE* wave) {
	if (trace_on)
		fprintf(stderr, "WaveSaveRaw()\n");

	if (fname == NULL || wave == NULL)
		return;
		
	FileAppend(fname, (UBYTE *)wave->samples, wave->size*4);
}



/*
 * Register WAVE. Add it to 'waves'. Added wave becomes selected one.
 *
 */
BOOL WaveRegister(WAVE* wave) {
	if (trace_on)
		fprintf(stderr, "WaveRegister(%s)\n", wave->name);
		
		
	if (nrwaves >= MAXWAVES) {
		fprintf(stderr, "Error: maximal number of waves reached\n");
		return(FALSE);
	}
			
	waves[nrwaves] = wave;
	idx = nrwaves;
	nrwaves++;
	
	WaveDisplay();
	return(TRUE);
}



/*
 * Play selected WAVE.
 *
 */
void WavePlaySelected(void) {
	INT i;
	ULONG nr_chunks;

	if (trace_on)
		fprintf(stderr, "WavePlaySelected()\n");
		
			
	if (nrwaves < 1)
		return;

	nr_chunks = waves[idx]->size/AudioGetBufferSize()+1;
	waves[idx]->position = 0;
	for (i = 0; i < nr_chunks; i++) {
		WavePlay(waves[idx]);
	}
}

/*
 * Plays WAVE. 
 *
 */
SWORD* WavePlay(WAVE* wave) {
	SWORD* audiobuffer;
	ULONG buffersize;
	
	if (trace_on)
		fprintf(stderr, "WavePlay(%s) wave->position=%li\n", wave->name, wave->position);

	if (wave == NULL)
		return(NULL);
		
	if (AudioOpened() == FALSE) {
		if (AudioOpen() == FALSE)
			return(NULL);
	}
	
	audiobuffer = AudioGetBuffer();
	buffersize = AudioGetBufferSize();

	// enough from wave to fill buffer?
	// yes
	if (wave->position + buffersize < wave->size) {
		WaveToRaw(audiobuffer, wave, wave->position, buffersize);
		AudioWrite(buffersize);
		wave->position = wave->position + buffersize;
	}
	else {
		WaveToRaw(audiobuffer, wave, wave->position, wave->size - wave->position);
		AudioWrite(wave->size - wave->position);
		wave->position = 0;
	}
	return(audiobuffer);
}


/*
 * Copy a block of size 'dest->size' from 'waves[idx]' to 'dest'. If
 * 'dest->size' >> 'waves[idx]->size', copy source until 'dest' full.
 * Returns index of following block of 'waves[idx]'
 *
 */
/*
ULONG WaveBlockToWave(WAVE* dest, WAVE* src, ULONG src_start, BOOL wraparound) {
	ULONG i, j = 0;
	
	if (trace_on)
		fprintf(stderr, "WaveBlockToWave()\n");

	
	if (src == NULL || dest == NULL)
		return(-1);

	for (i = 0; i < dest->size; i++) {
		j = (i+src_start) % src->size;
		dest->samples[i] = src->samples[j];
//fprintf(stderr, "j=%li, i=%li, smp=%f", j, i, dest->samples[i].left);
	}
	
	return(j+1);
}
*/


/*
 * Copy a SAMPLEs block of 'size' from 'src' to 'dest'. Loop over
 * 'src' until 'dest' full.
 * Returns index of following block in 'src'
 *
 */
ULONG WaveBlockToSample(SAMPLE* dest, WAVE* src, ULONG start, ULONG size, BOOL wraparound) {
	ULONG i;
	
	if (trace_on)
		fprintf(stderr, "WaveBlockToSample()\n");

	
	if (src == NULL) {
		fprintf(stderr, "WaveBlockToSample: src == NULL\n");
		return(-1);
	}
	if (dest == NULL) {
		fprintf(stderr, "WaveBlockToSample: dest == NULL\n");
		return(-1);
	}
	if (start < 0)
		return(-1);

	
	for (i = 0; i < size; i++, start++) {
		if (start == src->size && wraparound)
			start = 0;
		dest[i] = src->samples[start];
	}
//fprintf(stderr, "start=%li, i=%li", start, i);

	return(start);
}


/*
 * Clears 'samples' and 'buffer' of WAVE
 *
 */
void WaveClear(WAVE* wave) {
	ULONG buffer_size = AudioGetBufferSize();
	ULONG i;
	
	if (trace_on)
		fprintf(stderr, "WaveClearBuffer()\n");
            
	if (wave == NULL) {
		fprintf(stderr, "WaveClearBuffer: NULL ptr\n");
		return;
	}
	
	for (i = 0; i < wave->size; i++) {
		wave->samples[i].left = 0.0;
		wave->samples[i].right = 0.0;
	}
	for (i = 0; i < buffer_size; i++) {
		wave->buffer[i].left = 0.0;
		wave->buffer[i].right = 0.0;
	}
}


/*
 * Convert part of selected WAVE to raw. If raw buffer is bigger than WAVE,
 * leave rest of raw buffer untouched (maybe should be filled with 0)
 * IN: raw buffer, source wave, start position in 'src' and size in SAMPLEs
 * OUT: number of samples(==2 SWORDs) written to raw buffer
 *
 */
void WaveToRaw(SWORD* raw, WAVE* wave, ULONG start, ULONG size) {
	ULONG i;					// number of samples written
	FLOAT smpl;
	
	if (trace_on)
		fprintf(stderr, "WaveToRaw(%p, %li, %li)\n", &raw, start, size);

	if (wave == NULL)
		return;
		
	for (i = 0; i < size; i++) {
		if (start+i >= wave->size)
			break;
		smpl = wave->samples[start+i].left * 32767.0 * wave->lvol;
		if (smpl > 32767.0 && left_up_cut)
			smpl =  32767.0;
		if (smpl < -32768.0 && left_down_cut)
			smpl = -32768.0;
		raw[i*2] = (SWORD) smpl;
		smpl = wave->samples[start+i].right * 32767.0 * wave->rvol;
		if (smpl > 32767.0 && right_up_cut)
			smpl =  32767.0;
		if (smpl < -32768.0 && right_down_cut)
			smpl = -32768.0;
		raw[i*2+1] = (SWORD) smpl;
	}
}
	

/*
 * Allocates a new WAVE.
 * IN: name of wave, size in SAMPLEs
 */
WAVE* WaveNew(CHAR* name, ULONG size) {
	WAVE* wave;
	
		
	if (trace_on)
		fprintf(stderr, "WaveNew(\"%s\",%li)\n", name, size);


	wave = (WAVE*)malloc(sizeof(WAVE));
	if (wave == NULL)
		return NULL;
	strncpy(wave->name, name, 41);
	wave->size = size;
	wave->start = 0;
	wave->position = 0;
	wave->samples = (SAMPLE*)calloc(size, sizeof(SAMPLE));
	if (wave->samples == NULL)
		return(NULL);
	wave->lvol = 1.0;
	wave->rvol = 1.0;
	wave->lpan = 1.0;
	wave->rpan = 1.0;
	wave->buffer = (SAMPLE*)calloc(AudioGetBufferSize(), sizeof(SAMPLE));
	if (wave->buffer == NULL)
		return(NULL);
	wave->active = FALSE;
//fprintf(stderr, "wave->name=%s\n", wave->name);

	return(wave);
}


/*
 * Clone selected WAVE
 *
 */
WAVE* WaveClone(WAVE* wave) {
	WAVE* res;
	
	if (trace_on)
		fprintf(stderr, "WaveClone()\n");	

	if (wave == NULL) {
		fprintf(stderr, "WaveClone: NULL ptr!");
		return(NULL);
	}
	
	res = WaveNew(wave->name, wave->size);
	WaveCopy(res, wave);
	
	return(res);
}

	

/*
 * Copies 'src' to 'dest'. 'dest' retains its size. If 'dest' is larger than
 * 'src', the rest is filled with zeros.
 *
 */
WAVE* WaveCopy(WAVE* dest, WAVE* src) {
	INT i;

	if (trace_on)
		fprintf(stderr, "WaveCopy()\n");


	if (dest == NULL || src == NULL) {
		fprintf(stderr, "WaveCopy: NULL ptr!");
		return(NULL);
	}
	
	for (i = 0; i < dest->size; i++) {
		if (i >= src->size)
			break;
		dest->samples[i] = src->samples[i];
	}
	for (; i < dest->size; i++) {
		dest->samples[i].left = 0.0;
		dest->samples[i].right = 0.0;
	}

	dest->peak = src->peak;
	dest->size = src->size;
	dest->start = src->start;
	dest->position = src->position;
	dest->active = src->active;
	
	return(dest);
}




/*
 * Free WAVE's memory
 *
 */
void WaveFree(WAVE* wave) {
	if (trace_on)
		fprintf(stderr, "WaveFree()\n");

	if (wave == NULL )
		return;

	free(wave->samples);
	free(wave->buffer);
	free(wave);
}


/*
 * Delete 'waves[idx]' from 'waves' and destroy it
 *
 */
void WaveDelete(void) {
	if (trace_on)
		fprintf(stderr, "WaveDelete()\n");
		
	if (nrwaves < 1)
		return;

	WaveFree(waves[idx]);
	WaveSwap(nrwaves-1);
	nrwaves--;
	if (idx >= nrwaves)
		idx = nrwaves-1;
	if (idx < 0)
		idx = 0;
}




/*
 * Resize WAVE.
 *
 */
void WaveResize(ULONG newsize) {
	WAVE* wave;

	if (trace_on)
		fprintf(stderr, "WaveResize(%li)\n", newsize);


	if (nrwaves < 1)
		return;
		
	// right now it reallocates memory and copies whole sample: a waste
	if (newsize < waves[idx]->size)
		waves[idx]->size = newsize;
	else {
		wave = WaveNew(waves[idx]->name, newsize);
		memcpy(wave->samples, waves[idx]->samples, waves[idx]->size*sizeof(SAMPLE));
		wave->peak = waves[idx]->peak;
		wave->active = waves[idx]->active;
		if (waves[idx]->start > wave->size)
			wave->start = 0;
		else
			wave->start = waves[idx]->start;
		if (waves[idx]->position > wave->size)
			wave->position = 0;
		else
			wave->position = waves[idx]->position;
		WaveFree(waves[idx]);
		waves[idx] = wave;		
	}
}


/* 
 * Roll WAVE n SAMPLEs
 *
 */
void WaveRoll(SLONG offset) {
	WAVE* wave;
	ULONG i, j;
	
	if (trace_on)
		fprintf(stderr, "WaveRollLeft(%li)\n", offset);

	if (nrwaves < 1)
		return;


	if (abs(offset) > waves[idx]->size)
		offset = offset % waves[idx]->size;

	if (offset < 0)
		offset = waves[idx]->size+offset;

	if (offset == 0)
		return;
	// now we make a whole copy of wave. when using BREAKs just the 'BREAK->start'
	// has to be changed
	wave = WaveNew(waves[idx]->name, waves[idx]->size);
	wave->peak = waves[idx]->peak;
	wave->active = waves[idx]->active;
	wave->position = waves[idx]->position;
	wave->start = waves[idx]->start;
	
	// memcpy() should be faster!
	for (i = 0, j = offset; j < wave->size; i++, j++) {
		wave->samples[i] = waves[idx]->samples[j];
//fprintf(stderr, "sh1=%f ", wave->samples[i].left);
	}
	for (j = 0; i < wave->size; i++, j++) {
		wave->samples[i] = waves[idx]->samples[j];
//fprintf(stderr, "sh2=%f ", wave->samples[i].left);
	}
	WaveFree(waves[idx]);
	waves[idx] = wave;
}


	
/*
 * Normalize selected WAVE by 'WAVE->peak'
 * IN: WAVE*
 */
void WaveNormalize(void) {
	WAVE* wave = waves[idx];
	ULONG i;


	if (trace_on)
		fprintf(stderr, "WaveNormalize(%s, %f)\n", wave->name, wave->peak);

	
	if (nrwaves < 1)
		return;
		
	if (wave->peak == 0)	
		WavePeak();
	if (wave->peak == 0)
		return;

	for (i = 0; i < wave->size; i++) {
		wave->samples[i].left = wave->samples[i].left/wave->peak;
		wave->samples[i].right = wave->samples[i].right/wave->peak;
	}

	wave->peak = 1.0;
	
	WaveDisplay();
}


/*
 * Scale WAVE to 'amp'
 *
 */
void WaveScale(FLOAT amp) {
	WAVE* wave = waves[idx];
	ULONG i;


	if (trace_on)
		fprintf(stderr, "WaveScale(%s, %f)\n", wave->name, wave->peak);

	
	if (nrwaves < 1)
		return;
		
	WavePeak();
	if (wave->peak == 0)
		return;

	for (i = 0; i < wave->size; i++) {
		wave->samples[i].left = wave->samples[i].left*amp;
		wave->samples[i].right = wave->samples[i].right*amp;
	}

	wave->peak = wave->peak*amp;
	
	WaveDisplay();
}
	
/*
 * Absolute peak value of wave
 *
 */
FLOAT WavePeak(void) {
	WAVE* wave = waves[idx];
	ULONG i;
	FLOAT smp;


	if (trace_on)
		fprintf(stderr, "WavePeak(%s)\n", wave->name);

				
	if (nrwaves < 1)
		return(0);
		
	wave->peak = 0;
	for (i = 0; i < wave->size; i++) {
		smp = wave->samples[i].left;
		if (smp < 0)
			smp = -smp;
		if (smp > wave->peak)
			wave->peak = smp;
		smp = wave->samples[i].right;
		if (smp < 0)
			smp = -smp;
		if (smp > wave->peak)
			wave->peak = smp;
	}
	return(wave->peak);
}


WAVE* WaveNext(void) {
	if (trace_on)
		fprintf(stderr, "WaveNext()\n");
		
		
	if (idx < nrwaves-1)
		idx++;
	WaveDisplay();
	
	return(waves[idx]);
}


WAVE* WavePrevious(void) {
	if (trace_on)
		fprintf(stderr, "WavePrevious()\n");
		

	if (idx > 0)
		idx--;
	WaveDisplay();

	return(waves[idx]);
}


/*
 * select WAVE by name
 *
 */
void WaveSelectByName(CHAR* name) {
	INT i = WaveIdxByName(name);


	if (trace_on)
		fprintf(stderr, "WaveByName(%s)\n", name);
		
			
	if (i != -1)
		idx = i;
}


/*
 * select WAVE by number
 *
 */
void WaveSelectByNumber(INT number) {
	if (trace_on)
		fprintf(stderr, "WaveByNumber(%d)\n", number);


	if (number >= 0  &&  number < nrwaves)
		if (waves[number] != NULL)
			idx = number;
}


/*
 * returns selected WAVE
 *
 */
WAVE* WaveGet(void) {
	if (trace_on)
		fprintf(stderr, "WaveGet()\n");

	if (nrwaves < 1)
		return(NULL);
		
	return(waves[idx]);
}


/*
 * returns index ('idx') of selected wave. -1 if none available
 *
 */
INT WaveGetIndex(void) {
	if (trace_on)
		fprintf(stderr, "WaveGetIndex()\n");
	
	
	if (nrwaves < 1)
		return(-1);
	return(idx);
}


/*
 * returns WAVE by name
 *
 */
WAVE* WaveGetByName(CHAR* name) {
	INT i = WaveIdxByName(name);
	
	
	if (trace_on)
		fprintf(stderr, "WaveGetByName(%s)\n", name);
		
		
	if (i != -1)
		return(waves[i]);

	return(NULL);
}



/*
 * returns WAVE by its index in 'waves[]'
 *
 */
WAVE* WaveGetByIndex(INT number) {
	if (trace_on)
		fprintf(stderr, "WaveGetByIndex(%d)\n", number);
	
	
	if (number >= 0  &&  number < nrwaves)
		if (waves[number] != NULL)
			return(waves[number]);
			
	return(NULL);
}



void WaveDisplay(void) {
	if (trace_on)
	    fprintf(stderr, "WaveDisplay()\n");


	WaveClearDisplay();

	if (waves[idx] == NULL || nrwaves < 1) {
		mvprintw(1, 1, "No waves loaded");
		return;
	}
	
	mvprintw(0, 0, "wave:%s", waves[idx]->name);
	mvprintw(1, 0, "size:%li", waves[idx]->size);
	mvprintw(2, 0, "lvol:%f", waves[idx]->lvol);
	mvprintw(3, 0, "rvol:%f", waves[idx]->rvol);
	mvprintw(4, 0, "peak:%f", waves[idx]->peak);
	mvprintw(5, 0, "start:%li", waves[idx]->start);
	mvprintw(6, 0, "active:%d", waves[idx]->active);
}


void WaveClearDisplay(void) {
	if (trace_on)
		fprintf(stderr, "WaveClearDisplay()\n");


	mvprintw(0, 0, "                                         ");
	mvprintw(1, 0, "                                       ");
	mvprintw(2, 0, "                                     ");
	mvprintw(3, 0, "                                    ");
	mvprintw(4, 0, "                                   ");
	mvprintw(5, 0, "                            ");
}



INT WaveHowMany() {
	return(nrwaves);
}

	
//### PRIVATE METHODS ###########################################



/*
 * returns index of WAVE with name or -1
 *
 */
INT WaveIdxByName(CHAR* name) {
	INT i;
	
	
	if (trace_on)
		fprintf(stderr, "WaveIdxByName(%s)\n", name);


	for (i = 0; i < nrwaves; i++) {
		if (strcmp(name, waves[i]->name) == 0)
			return(i);
	}
	return(-1);
}


/*
 * Swap two WAVEs
 *
 */
void WaveSwap(INT sampleidx) {
	WAVE* tmp;
	
	
	if (trace_on)
 		fprintf(stderr, "WaveSwap(%d)\n", sampleidx);
 	
	
 	if (sampleidx >= nrwaves)
 		return;
	tmp = waves[idx];
	waves[idx] = waves[sampleidx];
	waves[sampleidx] = tmp;
}		
